#include <xsTypes.h>
#include <functionMap.h>
#include <stlToCArrays.h>
#include <XSUtil/Utils/XSutility.h>
#include <cfortran.h>
#include <algorithm>

// Functions called FROM here to Fortran:

PROTOCCALLSFSUB2(INITNEI,initnei,PINT,PINT)
#define INITNEI(nionp,nzmax) CCALLSFSUB2(INITNEI,initnei,PINT,PINT,nionp,nzmax)

PROTOCCALLSFSUB19(DOSHOCK,doshock,INT,FLOATV,INT,FLOATV,INT,FLOATV,FLOATV,FLOATV,
       INT,FLOATV,FLOATV,FLOAT,FLOAT,FLOAT,FLOAT,INT,PINT,INTV,INTV)
#define DOSHOCK(nt,delarr,nionp,ei,ntemp,eisum,t,fout,ntempr, tempr, tau, ts, taul, \
         tau0, beta,nzmax, ntmp, ionel, ionstage) \
        CCALLSFSUB19(DOSHOCK,doshock,INT,FLOATV,INT,FLOATV,INT,FLOATV,FLOATV,FLOATV, \
         INT,FLOATV,FLOATV,FLOAT,FLOAT,FLOAT,FLOAT,INT,PINT,INTV,INTV, \
         nt,delarr,nionp,ei,ntemp,eisum,t,fout,ntempr, tempr, tau, ts, taul, \
         tau0, beta,nzmax, ntmp, ionel, ionstage)

PROTOCCALLSFSUB16(NEISPEC,neispec,INT,FLOATV,INT,FLOATV,INTV,INTV,INT,FLOATV,FLOAT,FLOATV,INT,FLOATV,INTV,LOGICAL,STRING,FLOATV)
#define NEISPEC(ntemp,temp,nion,ionfrac,ionel,ionstage,ne,ear,z,linesh,nabun,abund,abunel,qcalc,callr,photar) \
           CCALLSFSUB16(NEISPEC,neispec,INT,FLOATV,INT,FLOATV,INTV,INTV,INT,FLOATV,FLOAT,FLOATV,INT,FLOATV,INTV,LOGICAL,STRING,FLOATV, \
           ntemp,temp,nion,ionfrac,ionel,ionstage,ne,ear,z,linesh,nabun,abund,abunel,qcalc,callr,photar) 


void xshock(const RealArray& energyArray, const RealArray& params,
        int spectrumNumber, RealArray& flux, RealArray& fluxErr, 
        const string& initString)
{
   //     Aug 23,99 subroutine hunt used instead of ibine for distribution of
   //               lines into bins.
   //     Aug 99    Plane-parallel shock models with unequal ion and electron 
   //               temperatures. May be used to model a shock section.
   //    	    additive model for XSPEC to use with general models.
   //		    parameters (param):
   //               (1):    postshock temperature t(keV)
   //               (2):    postshock electron temperature t_e(keV)
   //		    (3):    heavy metal abundance (with respect to solar)
   //               (4):    ionization parameter taul(cm^-3 s)
   //               (5):    ionization parameter tauu(cm^-3 s)
   //		    (6):    redshift z
   
   //   Translated to C++ from xsshock.f 03/09
   
   RealArray vparam(18);
   vparam[0] = params[0];
   vparam[1] = params[1];
   vparam[2] = 1.0;
   vparam[3] = 1.0;
   for (size_t i=4; i<=14; ++i)
      vparam[i] = params[2];
   vparam[15] = params[3];
   vparam[16] = params[4];
   vparam[17] = params[5];   

   // Call the xsshock subroutine for variable abundances.
   
   xsshock(energyArray, vparam, spectrumNumber, flux, fluxErr, initString);
}

void xsshock(const RealArray& energyArray, const RealArray& params,
        int spectrumNumber, RealArray& flux, RealArray& fluxErr, 
        const string& initString)
{
   //     Aug 99    Plane-parallel shock models with unequal ion and electron 
   //               temperatures. May be used to model a shock section.
   //		    additive model for XSPEC to use with general models.
   //		    parameters (param):
   //               (1):    postshock temperature t(keV)
   //               (2):    postshock electron temperature t_e(keV)
   //               (3):    hydrogen density in cm^-3
   //		    (4..15): abundances of He,C,N,O,Ne,Mg,Si,S,Ar,Ca,Fe,Ni
   //                                  with respect to solar values
   //               (16):   ionization parameter taul(cm^-3 s)
   //               (17):   ionization parameter tauu(cm^-3 s)
   //		    (18):   redshift z
   //
   //               Parameter (3) should be usually set to one. Its only use
   //               in the code is to set hydrogen abundance to zero; in this
   //               case this parameter should be set to zero. It is not the
   //               density parameter.
   
   using namespace XSutility;
   
   // the number of temperature values in the ionization tables.
   const int nt = 1251;
   const int ntemp = 300;
   float delarr[nt];
   float temp[ntemp];
   static float t[ntemp];
   
   static bool isFirst = true;
   static int nionp=0, nzmax=0;
   static auto_array_ptr<float> apEi(0);
   static auto_array_ptr<float> apEisum(0);
   static auto_array_ptr<float> apFout(0);
   static auto_array_ptr<int> apIonel(0);
   static auto_array_ptr<int> apIonstage(0);
   if (isFirst)
   {
      INITNEI(nionp, nzmax);
      // Don't need to delete the returned pointers since this can
      // only be entered once.
      apEi.reset(new float[nionp*nt]);
      apEisum.reset(new float[nionp*ntemp]);
      apFout.reset(new float[nionp]);
      apIonel.reset(new int[nionp]);
      apIonstage.reset(new int[nionp]);      
      isFirst = false;
   }
   float* ei = apEi.get();
   float* eisum = apEisum.get();
   float* fout = apFout.get();
   int* ionel = apIonel.get();
   int* ionstage = apIonstage.get();
   
   // conversion factors
   const float keV = 1.60219e-9;
   const float boltz = 1.38062e-16;
   const float tconv = keV/boltz;
   
   const float ts = tconv*static_cast<float>(std::max(params[0],params[1]));
   const float tse = tconv*static_cast<float>(std::min(params[0],params[1]));
   const float beta = tse/ts;
   const int nabun = 13;
   const float taul = static_cast<float>(std::min(params[nabun+2],params[nabun+3]));
   const float tau0 = static_cast<float>(std::max(params[nabun+2],params[nabun+3]));
   const float z = static_cast<float>(params[nabun+4]);
   
   const int ntempr = 200;
   float tempr[ntempr];
   float tau[ntempr];
   static float tprevs=0.0, tau0p=0.0, taulp=0.0, tsep=0.0;
   static int ntmp=0;
   
   bool doCalc = false;
   if (ts != tprevs || tau0 != tau0p || tse != tsep || taul != taulp)
   {
      // Ionic concentrations for solar abundances.
      // Zero out ionization fractions and emission measures.
      DOSHOCK(nt, delarr, nionp, ei, ntemp, eisum, t, fout, ntempr, tempr,
            tau, ts, taul, tau0, beta, nzmax, ntmp, ionel, ionstage);
      tprevs = ts;
      taulp = taul;
      tau0p = tau0;
      tsep = tse;
      doCalc = true;
   }

   // Calculate spectrum - need to convert temperatures to keV
   for (size_t it=0; it<ntmp; ++it)
      temp[it] = t[it]/tconv;
   
   static const int atomZ[nabun]={1, 2, 6, 7, 8, 10, 12, 14, 16, 18, 20, 26, 28};   
   int abunel[nabun];
   
   for (size_t it=0; it<nabun; ++it)
      abunel[it] = atomZ[it];
      
   int ne = energyArray.size()-1; 
   float linesh[2]={.0,.0}; 
   char funcName[] = "xsshock"; 
   float *ear=0, *pars=0, *photar=0, *photer=0;
   XSFunctions::stlToFloatArrays<float>(energyArray, params, flux, fluxErr,
           ear, pars, photar, photer);
   auto_array_ptr<float> apEar(ear);
   auto_array_ptr<float> apPars(pars);
   auto_array_ptr<float> apPhotar(photar);
   auto_array_ptr<float> apPhoter(photer);
   
   NEISPEC(ntmp, temp, nionp, eisum, ionel, ionstage, ne, ear, z, linesh,
           nabun, &pars[2], abunel, doCalc, funcName, photar);
           
   XSFunctions::floatFluxToStl<float>(photar, photer, ne, false, flux, fluxErr);       
   
}
